import networkx as nx
import pandas as pd
import numpy as np
import copy
from collections import Counter

from .genetic_routing import GeneticPaymentRouter

def get_shortest_paths(init_capacities, G_origi, transactions, hash_transactions=True, cost_prefix="", weight="total_fee", required_length=None):
    G = G_origi.copy()# copy due to forthcoming graph capacity changes!!!
    capacity_map = copy.deepcopy(init_capacities)
    with_depletion = capacity_map != None
    shortest_paths = []
    total_depletions = dict()
    router_fee_tuples = []
    hashed_transactions = {}
    genetic_rounds = []
    for idx, row in transactions.iterrows():
        p, cost = [], None
        try:
            S, T = row["source"], row["target"] + "_trg"
            if (not S in G.nodes()) or (not T in G.nodes()):
                shortest_paths.append((row["transaction_id"], cost, len(p)-1, p))
                continue
            p = nx.shortest_path(G, source=S, target=T, weight=weight)
            if required_length != None:
                if len(p) > 2 and len(p)-1 < required_length:
                    # extend only non-direct short chanels!
                    gpr = GeneticPaymentRouter(required_length, G)#, router_weights)
                    _, _, p_new, num_rounds = gpr.run(p, size=100, best_ratio=0.25)
                    genetic_rounds.append(num_rounds)
                    if num_rounds != -1:
                        p = p_new
            if row["target"] in p:
                raise RuntimeError("Loop detected: %s" % row["target"])
            cost, router_fees, depletions = process_path(p, row["amount_SAT"], capacity_map, G,  "total_fee", with_depletion)
            if with_depletion:
                for dep_node in depletions:
                    total_depletions[dep_node] = total_depletions.get(dep_node, 0) + 1
            routers = list(router_fees.keys())
            router_fee_tuples += list(zip([row["transaction_id"]]*len(router_fees),router_fees.keys(),router_fees.values()))
            if hash_transactions:
                for router in routers:
                    if not router in hashed_transactions:
                        hashed_transactions[router] = []
                    hashed_transactions[router].append(row)
        except nx.NetworkXNoPath:
            continue
        except:
            raise
        finally:
            shortest_paths.append((row["transaction_id"], cost, len(p)-1, p))
    if hash_transactions:
        for node in hashed_transactions:
            hashed_transactions[node] = pd.DataFrame(hashed_transactions[node], columns=transactions.columns)
    elif required_length!=None:
        cnt = Counter(genetic_rounds)
        print(cnt.most_common())
    all_router_fees = pd.DataFrame(router_fee_tuples, columns=["transaction_id","node","fee"])
    return pd.DataFrame(shortest_paths, columns=["transaction_id", cost_prefix+"cost", "length", "path"]), hashed_transactions,  all_router_fees, total_depletions

def process_path(path, amount_in_satoshi, capacity_map, G, weight, with_depletion):
    routers = {}
    depletions = []
    N = len(path)
    for i in range(N-2):
        n1, n2 = path[i], path[i+1]
        routers[n2] = G[n1][n2][weight]
        if with_depletion:
            n2_removed = process_forward_edge(capacity_map, G, amount_in_satoshi, n1, n2)
            if n2_removed:
                depletions.append(n2)
            process_backward_edge(capacity_map, G, amount_in_satoshi, n2, n1)
    # last node in path is always a pseudo node
    n1, n2 = path[N-2], path[N-1].replace("_trg","")
    if with_depletion:
        n2_removed = process_forward_edge(capacity_map, G, amount_in_satoshi, n1, n2)
        if n2_removed:
            depletions.append(n2)
        process_backward_edge(capacity_map, G, amount_in_satoshi, n2, n1)
    return np.sum(list(routers.values())), routers, depletions

def process_forward_edge(capacity_map, G, amount_in_satoshi, src, trg):
    removed = False
    cap, fee, is_trg, total_cap = capacity_map[(src,trg)]
    if cap < amount_in_satoshi:
        raise RuntimeError("forward %i: %s-%s" % (cap,src,trg))
    if cap < 2*amount_in_satoshi: # cannot route more transactions
        removed = True
        G.remove_edge(src, trg)
        if is_trg:
            G.remove_edge(src, trg+'_trg')
    capacity_map[(src,trg)] = [cap-amount_in_satoshi, fee, is_trg, total_cap]
    return removed
    
def process_backward_edge(capacity_map, G, amount_in_satoshi, src, trg):
    if (src,trg) in capacity_map:
        cap, fee, is_trg, total_cap = capacity_map[(src,trg)]
        if cap < amount_in_satoshi: # it can route transactions again
            G.add_weighted_edges_from([(src,trg,fee)], weight="total_fee")
            if is_trg:
                G.add_weighted_edges_from([(src,trg+'_trg',0.0)], weight="total_fee")
        capacity_map[(src,trg)] = [cap+amount_in_satoshi, fee, is_trg, total_cap]



import networkx as nx
import numpy as np
import pandas as pd

def populate_route(route, k, G, router_weights=None):
    """Populate short routes with random neighbors"""
    path = route.copy()
    trial_cnt = 0
    max_trials = 10
    success = True
    target = route[-1].replace("_trg","")
    while len(path) < k+1:
        pos = np.random.choice(range(len(path)-1))
        n1, n2 = path[pos], path[pos+1]
        # find common directed neighbor
        neigh = set(G.successors(n1)).intersection(set(G.predecessors(n2)))
        if target in neigh:
            neigh.remove(target)
        # exclude loops
        neigh = list(neigh.difference(set(path)))
        if len(neigh) == 0:
            trial_cnt += 1
            if trial_cnt == max_trials:
                success = False
                break
            continue
        probas = None
        if router_weights != None:
            weights = np.array([float(router_weights.get(n, 0.0)) for n in neigh])
            sum_ = np.sum(weights)
            if sum_ > 0:
                 probas = weights / sum_
        new_node = np.random.choice(neigh, p=probas)
        path.insert(pos+1, new_node)
    # finalize the instance
    return success, tuple(path)

def calculate_cost(route, G):
    """Calculate transaction cost for the sender node"""
    s = 0.0
    # last edge has no routing cost
    for i in range(len(route)-2):
        n1, n2 = route[i], route[i+1]
        s += G[n1][n2]["total_fee"]
    return s

def validate_path(route, G):
    # all nodes are unique
    if len(set(route)) == len(route):
        # all links are present
        for i in range(len(route)-1):
            n1, n2 = route[i], route[i+1]
            valid = G.has_edge(n1, n2)
            if not valid:
                print("No edge: %s - %s" % (n1,n2))
                break
        return valid
    else:
        print("Node duplication!")
        return False 

def mix_routes(route_1, route_2, G):
    """Randomly select neighbors from other routes"""
    target = route_1[-1].replace("_trg","")
    res = []
    for i in range(1,len(route_1)-1):
        n1, n2 = route_1[i-1], route_1[i+1]
        neigh = set(G.successors(n1)).intersection(set(G.predecessors(n2)))
        neigh = neigh.difference(set(route_1))
        if target in neigh:
            neigh.remove(target)
        neigh = list(neigh.intersection(set(route_2[1:-1])))
        if len(neigh) > 0:
            r1 = list(route_1)
            r1[i] = np.random.choice(neigh)
            if not validate_path(r1, G):
                raise RuntimeError("Invalid path: %s" % r1)
            res.append(tuple(r1))
    return res

class GeneticPaymentRouter():
    def __init__(self, k, G, router_weights=None):
        self.k = k
        self.G = G
        self.router_weights = router_weights
        
    def _init_population(self, route, size):
        population = []
        for _ in range(size):
            success, path = populate_route(route, self.k, self.G, self.router_weights)
            if success:
                population.append(path)
        # remove duplications
        return list(set(population))
    
    def _eval_population(self, population):
        costs = [calculate_cost(item, self.G) for item in population]
        pop_df = pd.DataFrame(list(zip(population, costs)), columns=["route","cost"])
        opt_record = pop_df.nsmallest(1,"cost").values
        opt_path, opt_cost = opt_record[0,0], opt_record[0,1]
        return pop_df, opt_path, opt_cost
    
    def _gen_offsprings(self, pop_df, cnt, times=5):
        """Generate offsprings from best individuals. Additional random individuals are also sampled from the previous population"""
        individuals = list(pop_df["route"])
        N = len(individuals)
        parents = list(pop_df.nsmallest(cnt, "cost")["route"])
        L = len(parents)
        offsprings = []
        for _ in range(times):
            # random permutation
            np.random.shuffle(parents)
            # generate new offsprings pairwise
            for i in range(0,L-1,2):
                p1, p2 = parents[i], parents[i+1]
                offsprings += mix_routes(p1, p2, self.G)
                offsprings += mix_routes(p2, p1, self.G)
        unique_offsprings = list(set(offsprings))
        rnd_indices = set(np.random.choice(range(L), size=len(unique_offsprings), replace=True))
        return unique_offsprings + [individuals[j] for j in rnd_indices]
            
    def run(self, route, size=100, best_ratio=0.25, iterations=5, verbose=False):
        """Run fixed size minimal cost search with genetic algorithm"""
        pop = self._init_population(route, size)
        if len(pop) == 0:
            return calculate_cost(route, self.G), len(route)-1, route, -1
        else:
            pop_df, opt_path, opt_cost = self._eval_population(pop)
            if verbose:
                print("init", len(pop), opt_cost)
            for idx in range(iterations):
                pop = self._gen_offsprings(pop_df, int(size*best_ratio))
                if len(pop) == 0:
                    if verbose:
                        print("Empty population in %i round!" % (idx+1))
                    break
                pop_df, new_path, new_cost = self._eval_population(pop)
                if new_cost < opt_cost:
                    opt_cost = new_cost
                    opt_path = new_path
                else:
                    if verbose:
                        print("Early stopping in %i round!" % (idx+1))
                    break
                if verbose:
                    print(len(pop), new_cost)
            return opt_cost, len(opt_path)-1, opt_path, idx



import pandas as pd
import numpy as np

def sample_providers(node_variables, K, providers):
    provider_records = node_variables[node_variables["pub_key"].isin(providers)]
    nodes = list(provider_records["pub_key"])
    probas = list(provider_records["degree"] / provider_records["degree"].sum())
    return np.random.choice(nodes, size=K, replace=True, p=probas)

def sample_transactions(node_variables, amount_in_satoshi, K, eps, active_providers, verbose=False):
    nodes = list(node_variables["pub_key"])
    src_selected = np.random.choice(nodes, size=K, replace=True)
    if eps > 0:
        n_prov = int(eps*K)
        trg_providers = sample_providers(node_variables, n_prov, active_providers)
        trg_rnd = np.random.choice(nodes, size=K-n_prov, replace=True)
        trg_selected = np.concatenate((trg_providers,trg_rnd))
        np.random.shuffle(trg_selected)
    else:
        trg_selected = np.random.choice(nodes, size=K, replace=True)
    transactions = pd.DataFrame(list(zip(src_selected, trg_selected)), columns=["source","target"])
    transactions["amount_SAT"] = amount_in_satoshi
    transactions["transaction_id"] = transactions.index
    transactions = transactions[transactions["source"] != transactions["target"]]
    if verbose:
        print("Number of loop transactions (removed):", K-len(transactions))
        print("Merchant target ratio:", len(transactions[transactions["target"].isin(active_providers)]) / len(transactions))
    return transactions[["transaction_id","source","target","amount_SAT"]]


import networkx as nx
import pandas as pd
import numpy as np

def prepare_edges_for_simulation(edges, amount_sat, drop_disabled, drop_low_cap, time_window=None, ts_upper_bound=None, verbose=True):
    """Preprocess LN graph snapshot with different edge filters."""
    E = len(edges)
    if ts_upper_bound != None:
        tmp_edges = edges[edges["last_update"] < ts_upper_bound].copy()
    else:
        tmp_edges = edges.copy()
    if verbose:
        print("Channel filter - invalid timestamp:", E - len(tmp_edges))
    # remove edges with capacity below threshold
    if drop_low_cap:
        tmp_edges = tmp_edges[tmp_edges["capacity"] >= amount_sat]
        if verbose:
            print("Channel filter - capacity:", E - len(tmp_edges))
        E = len(tmp_edges)
    # remove old channels
    if time_window != None:
        ts_lower_bound = tmp_edges["last_update"].max() - time_window
        tmp_edges = tmp_edges[tmp_edges["last_update"] >= ts_lower_bound]
        if verbose:
            print("Channel filter - recency:", E - len(tmp_edges))
        E = len(tmp_edges)
    # remove disabled edges
    if drop_disabled:
        tmp_edges = tmp_edges[~tmp_edges["disabled"]]
        if verbose:
            print("Channel filter - disabled:", E - len(tmp_edges))
        E = len(tmp_edges)
    # calculate edge costs
    tmp_edges["total_fee"] = calculate_tx_fee(tmp_edges, amount_sat)
    # aggregate multi-edges
    grouped = tmp_edges.groupby(["src","trg"])
    directed_aggr_edges = grouped.agg({
        "capacity":"sum",
        "total_fee":"mean",
    }).reset_index()
    if verbose:
        print("Total number of deleted directed channels:", len(edges) - len(tmp_edges))
        print("Number of remaining directed channels:", len(tmp_edges))
        print("Number of edges after aggregation: %i" % len(directed_aggr_edges))
    return directed_aggr_edges

def init_node_params(edges, providers, verbose=True):
    """Initialize source and target distribution of each node in order to drawn transaction at random later."""
    G = nx.from_pandas_edgelist(edges, source="src", target="trg", edge_attr=["capacity"], create_using=nx.DiGraph())
    active_providers = list(set(providers).intersection(set(G.nodes())))
    active_ratio = len(active_providers) / len(providers)
    if verbose:
        print("Total number of possible providers: %i" % len(providers))
        print("Ratio of active providers: %.2f" % active_ratio)
    degrees = pd.DataFrame(list(G.degree()), columns=["pub_key","degree"])
    total_capacity = pd.DataFrame(list(nx.degree(G, weight="capacity")), columns=["pub_key","total_capacity"])
    node_variables = degrees.merge(total_capacity, on="pub_key")
    #get_src_rayleigh_proba(node_variables)
    #get_trg_proba(node_variables, eps=0.95, active_providers)
    return node_variables, active_providers, active_ratio

def generate_graph_for_path_search(edges, transactions, amount_sat):
    """Generate pseudo edges and generate graph for path search."""
    targets = list(transactions["target"].unique())
    #sources = list(transactions["source"].unique())
    #participants = set(sources).union(set(targets))
    # drop edges with low capacity
    edges_tmp = edges[edges["capacity"]>=amount_sat].copy()
    # add pseudo targets
    ps_edges = edges_tmp[edges_tmp["trg"].isin(targets)].copy()
    ps_edges["trg"] = ps_edges["trg"].apply(lambda x: str(x) + "_trg")
    ps_edges["total_fee"] = 0.0
    ps_edges["fee_base_msat"] = 0.0
    ps_edges["fee_rate_milli_msat"] = 0.0
    # initialize transaction graph
    all_edges = pd.concat([edges_tmp, ps_edges], sort=False)
    # networkx versiom >= 2: from_pandas_edgelist
    G = nx.from_pandas_edgelist(all_edges, source="src", target="trg", edge_attr=["total_fee","capacity"], create_using=nx.DiGraph())
    return G

def calculate_tx_fee(df, amount_sat):
    # first part: fee_base_msat -> fee_base_sat
    # second part: milli_msat == 10^-6 sat : fee_rate_milli_msat -> fee_rate_sat
    return df["fee_base_msat"] / 1000.0 + amount_sat * df["fee_rate_milli_msat"] / 10.0**6

def init_capacities(edges, transactions, amount_sat, verbose=False):
    """Initialize capacity map for path search"""
    tx_targets = set(transactions["target"])
    # init capacity dict
    keys = list(zip(edges["src"], edges["trg"]))
    is_trg = edges["trg"].apply(lambda x: x in tx_targets)
    # [current_cap, total_fee, is_trg, total_cap]
    vals = [list(item) for item in zip([None]*len(edges), edges["total_fee"], is_trg, edges["capacity"])]
    capacity_map = dict(zip(keys,vals))
    # extract channels
    channels = set()
    for s, t in keys:
        if (s,t) in channels or (t,s) in channels:
            continue
        else:
            channels.add((s,t))
    edges_with_capacity = populate_capacities(channels, capacity_map, amount_sat)
    if verbose:
        print("Edges with capacity: %i->%i" % (len(edges),len(edges_with_capacity))) 
    return capacity_map, edges_with_capacity
    
def populate_capacities(channels, capacity_map, amount_sat):
    """Initialize the capacity state of each channel at random"""
    edge_records = []
    for src, trg in channels:
        c1 = capacity_map[(src,trg)][3]
        if (trg,src) in capacity_map:
            c2 = capacity_map[(trg,src)][3]
            cap = max(c1, c2)
            rnd = np.random.random()
            cap1, cap2 = cap * rnd, cap * (1.0-rnd)
            capacity_map[(trg,src)][0] = cap2
            if cap2 >= amount_sat:
                edge_records.append((trg, src, cap2, capacity_map[(trg,src)][1]))
        else:
            cap1 = c1
        capacity_map[(src,trg)][0] = cap1
        if cap1 >= amount_sat:
            edge_records.append((src, trg, cap1, capacity_map[(src,trg)][1]))
    return pd.DataFrame(edge_records, columns=["src","trg","capacity","total_fee"])



import sys, os, json
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from tqdm import tqdm
import functools
import concurrent.futures

from .transaction_sampling import sample_transactions
from .graph_preprocessing import *
from .path_searching import get_shortest_paths

def shortest_paths_with_exclusion(capacity_map, G, cost_prefix, weight, hash_bucket_item):
    node, bucket_transactions = hash_bucket_item
    H = G.copy()
    H.remove_node(node)
    if node + "_trg" in G.nodes():
        H.remove_node(node + "_trg") # delete node copy as well
    new_paths, _, _, _ = get_shortest_paths(capacity_map, H, bucket_transactions,  hash_transactions=False, cost_prefix=cost_prefix, weight=weight)
    new_paths["node"] = node
    return new_paths

def get_shortest_paths_with_node_removals(capacity_map, G, hashed_transactions, cost_prefix="", weight=None, threads=4):
    print("Parallel execution on %i threads in progress.." % threads)
    if threads > 1:
        f_partial = functools.partial(shortest_paths_with_exclusion, capacity_map, G, cost_prefix, weight)
        executor = concurrent.futures.ProcessPoolExecutor(threads)
        alternative_paths = list(executor.map(f_partial, hashed_transactions.items()))
        executor.shutdown()
    else:
        alternative_paths = []
        for hash_bucket_item in tqdm(hashed_transactions.items(), mininterval=10):
            alternative_paths.append(shortest_paths_with_exclusion(capacity_map, G, cost_prefix, weight, hash_bucket_item))
    return pd.concat(alternative_paths)

class TransactionSimulator():
    def __init__(self, edges, merchants, amount_sat, count, epsilon=0.8, drop_disabled=True, drop_low_cap=True, with_depletion=True, time_window=None, verbose=False):
        self.verbose = verbose
        self.with_depletion = with_depletion
        self.amount = amount_sat
        self.edges = prepare_edges_for_simulation(edges, amount_sat, drop_disabled, drop_low_cap, time_window, verbose=self.verbose)
        self.node_variables, self.merchants, active_ratio = init_node_params(self.edges, merchants, verbose=self.verbose)
        self.transactions = sample_transactions(self.node_variables, amount_sat, count, epsilon, self.merchants, verbose=self.verbose)
        self.params = {
            "amount":amount_sat,
            "count":count,
            "epsilon":epsilon,
            "with_depletion":with_depletion,
            "drop_disabled":drop_disabled,
            "drop_low_cap": drop_low_cap,
            "time_window":time_window
        }
    
    def simulate(self, weight="total_fee", with_node_removals=False, max_threads=2, excluded=[], required_length=None, cap_change_nodes=[], capacity_fraction=1.0):
        edges_tmp = self.edges.copy()
        if len(cap_change_nodes) > 0 and capacity_fraction < 1.0:
            filt = edges_tmp["src"].isin(cap_change_nodes) | edges_tmp["trg"].isin(cap_change_nodes)
            edges_tmp.loc[filt,"capacity"] *= capacity_fraction
            edges_tmp = edges_tmp[edges_tmp["capacity"] >= self.amount]
            print("Capacity change executed: (%s, %.4f)" % (str(cap_change_nodes), capacity_fraction))
        if self.with_depletion:
            current_capacity_map, edges_with_capacity = init_capacities(edges_tmp, self.transactions, self.amount, self.verbose)
            G = generate_graph_for_path_search(edges_with_capacity, self.transactions, self.amount)
        else:
            current_capacity_map = None
            G = generate_graph_for_path_search(edges_tmp, self.transactions, self.amount)
        if len(excluded) > 0:
            print(G.number_of_edges(), G.number_of_nodes())
            for node in excluded:
                if node in G.nodes():
                    G.remove_node(node)
                pseudo_node = str(node) + "_trg"
                if pseudo_node in G.nodes():
                    G.remove_node(pseudo_node)
            if self.verbose:
                print(G.number_of_edges(), G.number_of_nodes())
            print("Additional nodes were EXCLUDED!")
        print("Graph and capacities were INITIALIZED")
        if self.verbose:
            print("Using weight='%s' for the simulation" % weight)    
        print("Transactions simulated on original graph STARTED..")
        shortest_paths, hashed_transactions, all_router_fees, total_depletions = get_shortest_paths(current_capacity_map, G, self.transactions, hash_transactions=with_node_removals, cost_prefix="original_", weight=weight, required_length=required_length)
        success_tx_ids = set(all_router_fees["transaction_id"])
        self.transactions["success"] = self.transactions["transaction_id"].apply(lambda x: x in success_tx_ids)
        print("Transactions simulated on original graph DONE")
        print("Transaction succes rate:")
        print(self.transactions["success"].value_counts() / len(self.transactions))
        if self.verbose:
            print("Length distribution of optimal paths:")
            print(shortest_paths["length"].value_counts())
        if with_node_removals:
            print("Base fee optimization STARTED..")
            alternative_paths = get_shortest_paths_with_node_removals(current_capacity_map, G, hashed_transactions, weight=weight, threads=max_threads)
            print("Base fee optimization DONE")
            if self.verbose:
                if verbose:
                    print("Length distribution of optimal paths:")
                    print(alternative_paths["length"].value_counts())
        else:
            alternative_paths = pd.DataFrame([])
        self.shortest_paths = shortest_paths
        self.alternative_paths = alternative_paths
        self.all_router_fees = all_router_fees
        return shortest_paths, alternative_paths, all_router_fees, total_depletions
    
    def export(self, output_dir):
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        with open('%s/params.json' % output_dir, 'w') as fp:
            json.dump(self.params, fp)
        length_distrib = self.shortest_paths["length"].value_counts()
        length_distrib.to_csv("%s/lengths_distrib.csv" % output_dir)
        total_income = get_total_income_for_routers(self.all_router_fees)
        total_income.to_csv("%s/router_incomes.csv" % output_dir, index=False)
        total_fee = get_total_fee_for_sources(self.transactions, self.shortest_paths)
        total_fee.to_csv("%s/source_fees.csv" % output_dir, index=True)
        print("Export DONE")
        return total_income, total_fee
    
### process results ###

def get_total_income_for_routers(all_router_fees):
    grouped = all_router_fees.groupby("node")
    aggr_router_income = grouped.agg({"fee":"sum","transaction_id":"count"}).reset_index().sort_values("fee",ascending=False)
    return aggr_router_income.rename({"transaction_id":"num_trans"}, axis=1)

def get_total_fee_for_sources(transactions, shortest_paths):
    tmp_sp = shortest_paths[shortest_paths["length"]>0]
    trans_with_costs = transactions[["transaction_id","source"]].merge(tmp_sp[["transaction_id","original_cost"]], on="transaction_id", how="right")
    #agg_funcs = dict(original_cost='mean', transaction_id='count')
    agg_funcs = {"original_cost":'mean', "transaction_id":'count'}
    aggs = trans_with_costs.groupby("source").agg(agg_funcs).rename({"original_cost":"mean_fee","transaction_id":"num_trans"}, axis=1)
    return aggs

### optimal fee pricing ###

def inspect_base_fee_thresholds(ordered_deltas, pos_thresholds, min_ratio):
    thresholds = [0.0] + pos_thresholds
    original_income = ordered_deltas["fee"].sum()
    original_num_transactions = len(ordered_deltas)
    incomes, probas = [original_income], [1.0]
    # inspect only positive deltas
    for th in thresholds[1:]:
        # transactions that will still pay the increased base_fee
        df = ordered_deltas[ordered_deltas["delta_cost"] >= th]
        prob = len(df) / original_num_transactions
        probas.append(prob)
        # adjusted router income at the new threshold
        adj_income = df["fee"].sum() + len(df) * th
        incomes.append(adj_income)
        if prob < min_ratio:
            break
    return incomes, probas, thresholds, original_income, original_num_transactions

def visualize_thresholds(incomes, probas, thresholds, original_num_transactions):
    fig, ax1 = plt.subplots()
    x = thresholds[:len(incomes)]
    ax1.set_title(original_num_transactions)
    ax1.plot(x, incomes, 'bx-')
    ax1.set_xscale("log")
    ax2 = ax1.twinx()
    ax2.plot(x, probas, 'gx-')
    ax2.set_xscale("log")

def calculate_max_income(n, p_altered, shortest_paths, all_router_fees, visualize=False, min_ratio=0.0):
    trans = p_altered[p_altered["node"] == n]
    trans = trans.merge(shortest_paths[["transaction_id","original_cost"]], on="transaction_id", how="inner")
    trans = trans.merge(all_router_fees, on=["transaction_id","node"], how="inner")#'fee' column is merged
    # router could ask for this cost difference
    trans["delta_cost"] = trans["cost"] - trans["original_cost"]
    ordered_deltas = trans[["transaction_id","fee","delta_cost"]].sort_values("delta_cost")
    ordered_deltas["delta_cost"] = ordered_deltas["delta_cost"].apply(lambda x: round(x, 2))
    pos_thresholds = sorted(list(ordered_deltas[ordered_deltas["delta_cost"]>0.0]["delta_cost"].unique()))
    incomes, probas, thresholds, alt_income, alt_num_trans = inspect_base_fee_thresholds(ordered_deltas, pos_thresholds, min_ratio)
    if visualize:
        visualize_thresholds(incomes, probas, thresholds, alt_num_trans)
    max_idx = np.argmax(incomes)
    return thresholds[max_idx], incomes[max_idx], probas[max_idx], alt_income, alt_num_trans

def calc_optimal_base_fee(shortest_paths, alternative_paths, all_router_fees):
    # paths with length at least 2
    valid_sp = shortest_paths[shortest_paths["length"]>1]
    # drop failed alternative paths
    p_altered = alternative_paths[~alternative_paths["cost"].isnull()]
    num_routers = len(alternative_paths["node"].unique())
    num_routers_with_alternative_paths = len(p_altered["node"].unique())
    routers = list(p_altered["node"].unique())
    opt_strategy = []
    for n in tqdm(routers, mininterval=5):
        opt_delta, opt_income, opt_ratio, origi_income, origi_num_trans = calculate_max_income(n, p_altered, valid_sp, all_router_fees, visualize=False)
        opt_strategy.append((n, opt_delta, opt_income, opt_ratio, origi_income, origi_num_trans))
    opt_fees_df = pd.DataFrame(opt_strategy, columns=["node", "opt_delta", "opt_alt_income", "opt_alt_traffic", "alt_income", "alt_traffic"])
    total_income = get_total_income_for_routers(all_router_fees).rename({"fee":"total_income","num_trans":"total_traffic"}, axis=1)
    merged_infos = total_income.merge(opt_fees_df, on="node", how="outer")
    merged_infos = merged_infos.sort_values("total_income", ascending=False)
    merged_infos = merged_infos.fillna(0.0)
    merged_infos["failed_traffic"] = merged_infos["total_traffic"] - merged_infos["alt_traffic"]
    merged_infos["failed_traffic_ratio"] = merged_infos["failed_traffic"] / merged_infos["total_traffic"]
    merged_infos["income_diff"] = merged_infos.apply(lambda x: x["opt_alt_income"] - x["alt_income"] +  x["failed_traffic"] * x["opt_delta"], axis=1)
    merged_infos
    #print(merged_infos.drop("node", axis=1).mean())
    return merged_infos[["node","total_income","total_traffic","failed_traffic_ratio","opt_delta","income_diff"]], p_altered



from setuptools import find_packages, setup

install_requires = [
    "pandas",
    "numpy",
    "networkx",
    "matplotlib",
    "seaborn",
    "tqdm",
]

#setup_requires = ['pytest-runner']

#tests_require = [
#    'pytest',
#    'pytest-cov',
#    'codecov'
#]

keywords = [
    "bitcoin",
    "lightning-network",
    "simulator",
    "simulation",
    "research",
    "cryptoeconomics"
]

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='lnsimulator',
      version='0.1.0',
      description="Traffic Simulator for Bitcoin's Lightning Network ",
      url='https://github.com/ferencberes/LNTrafficSimulator',
      author='Ferenc Beres',
      author_email='fberes@info.ilab.sztaki.hu',
      packages = find_packages(),
      install_requires=install_requires,
      #setup_requires = setup_requires,
      #tests_require = tests_require,
      keywords = keywords,
      long_description=long_description,
      long_description_content_type='text/markdown',
)



import json
from tqdm import tqdm
import pandas as pd

def load_temp_data(json_files, node_keys=["pub_key","last_update"], edge_keys=["node1_pub","node2_pub","last_update","capacity"]):
    """Load LN graph json files from several snapshots"""
    node_info, edge_info = [], []
    for idx, json_f in enumerate(json_files):
        with open(json_f) as f:
            try:
                tmp_json = json.load(f)
            except json.JSONDecodeError:
                print("JSONDecodeError: " + json_f)
                continue
        new_nodes = pd.DataFrame(tmp_json["nodes"])[node_keys]
        new_edges = pd.DataFrame(tmp_json["edges"])[edge_keys]
        new_nodes["snapshot_id"] = idx
        new_edges["snapshot_id"] = idx
        print(json_f, len(new_nodes), len(new_edges))
        node_info.append(new_nodes)
        edge_info.append(new_edges)
    edges = pd.concat(edge_info)
    edges["capacity"] = edges["capacity"].astype("int64")
    edges["last_update"] = edges["last_update"].astype("int64")
    print("All edges:", len(edges))
    edges_no_loops = edges[edges["node1_pub"] != edges["node2_pub"]]
    print("All edges without loops:", len(edges_no_loops))
    return pd.concat(node_info), edges_no_loops

def generate_directed_graph(edges, policy_keys=['disabled', 'fee_base_msat', 'fee_rate_milli_msat', 'min_htlc']):
    """Generate directed graph data from undirected payment channels."""
    directed_edges = []
    indices = edges.index
    for idx in tqdm(indices):
        row = edges.loc[idx]
        e1 = [row[x] for x in ["snapshot_id","node1_pub","node2_pub","last_update","channel_id","capacity"]]
        e2 = [row[x] for x in ["snapshot_id","node2_pub","node1_pub","last_update","channel_id","capacity"]]
        if row["node2_policy"] == None:
            e1 += [None for x in policy_keys]
        else:
            e1 += [row["node2_policy"][x] for x in policy_keys]
        if row["node1_policy"] == None:
            e2 += [None for x in policy_keys]
        else:
            e2 += [row["node1_policy"][x] for x in policy_keys]
        directed_edges += [e1, e2]
    cols = ["snapshot_id","src","trg","last_update","channel_id","capacity"] + policy_keys
    directed_edges_df = pd.DataFrame(directed_edges, columns=cols)
    return directed_edges_df

def preprocess_json_file(json_file):
    """Generate directed graph data (traffic simulator input format) from json LN snapshot file."""
    json_files = [json_file]
    print("\ni.) Load data")
    EDGE_KEYS = ["node1_pub","node2_pub","last_update","capacity","channel_id",'node1_policy','node2_policy']
    nodes, edges = load_temp_data(json_files, edge_keys=EDGE_KEYS)
    print(len(nodes), len(edges))
    print("Remove records with missing node policy")
    print(edges.isnull().sum() / len(edges))
    origi_size = len(edges)
    edges = edges[(~edges["node1_policy"].isnull()) & (~edges["node2_policy"].isnull())]
    print(origi_size - len(edges))
    print("\nii.) Transform undirected graph into directed graph")
    directed_df = generate_directed_graph(edges)
    #print(directed_df.head())
    print("\niii.) Fill missing policy values with most frequent values")
    print("missing values for columns:")
    print(directed_df.isnull().sum())
    directed_df = directed_df.fillna({"disabled":False,"fee_base_msat":1000,"fee_rate_milli_msat":1,"min_htlc":1000})
    for col in ["fee_base_msat","fee_rate_milli_msat","min_htlc"]:
        directed_df[col] = directed_df[col].astype("float64")
    return directed_df
    